# -*- coding: utf-8 -*-
import Function_for_application_test_python3
import pandas as pd
import glob
import os
import argparse

import numpy as np

parser = argparse.ArgumentParser()
parser.add_argument("-p", "--path", type=str, help="Path of the csv file")
parser.add_argument("-dp", "--debut_path", type=str, help="Path beginning of the csv (default: %(default)s voxels)",
                    default='')
parser.add_argument("-pa", "--patch", type=str, help="Patch size (default: %(default)s)", default=128)
parser.add_argument("-s", "--step", type=str, help="Step between patches. Must be a tuple of tuple (default: "
                                                   "%(default)s)", default=64)
parser.add_argument("-rf", "--result_folder_name", type=str, help='Name of the folder where the result is going to be '
                                                                  'stored')
parser.add_argument("-wp", "--weights_relative_path", type=str, help='Weights relative paths')
parser.add_argument("-bb", "--by_batch", type=bool, help="Prediction on list of patches instead of using a for loop. "
                                                         "Enables for instance to automatically computes in multi-gpu "
                                                         "mode(default: %(default)s)", default=False)
parser.add_argument("-ic", "--is_conditional", type=bool, help="Enable conditional prediction on z resolution")
args = parser.parse_args()

by_batch = args.by_batch
is_conditional = args.is_conditional
# Argument :
# name of the result folder
result_folder = args.result_folder_name

weights_relative_path = args.weights_relative_path

debut_relatif_path = args.debut_path  # Path to Base_pour_romeo

data = pd.read_csv(debut_relatif_path + args.path, header=None).iloc[:, 0]

path_pour_application = debut_relatif_path + data

path_pour_application = path_pour_application.sort_values()


def list_of_lists(arg):
    m = [x.split(' ') for x in arg.split(',')]
    m = np.array(m).astype('int')
    return m.tolist()


def list_of(arg):
    m = [x for x in arg.split(',')]
    m = np.array(m).astype('int')
    return m.tolist()


ensemble_pas = list_of_lists(args.step)
patchs = list_of(args.patch)

# Function :


def create_folder(directory):
    try:
        if not os.path.exists(directory):
            os.makedirs(directory)
    except OSError:
        print('Error: Creating directory. ' + directory)

# Execution :


ensemble_pas = pd.DataFrame(ensemble_pas, index=patchs)

np_pas_per_patch = ensemble_pas.shape[1]

weights_path = os.getcwd() + weights_relative_path

for i in path_pour_application:

    for patch in patchs:

        if patch is None:

            i_split = i.split("/")

            path_output = "/".join(i_split[:(len(i_split) - 1)]) + "/code avec " + result_folder + "/patch " + str(
                patch)

            print(path_output)

            if not os.path.exists(path_output):

                create_folder(path_output)

                path_output_cortex = path_output + "/Cortex " + str(patch) + ".nii.gz"

                path_output_SR = path_output + "/SR " + str(patch) + ".nii.gz"

                Function_for_application_test_python3.segmentation(input_file_path=i,
                                                                   step=20,
                                                                   NewResolution=(0.5, 0.5, 0.5),
                                                                   patch=patch,
                                                                   path_output_cortex=path_output_cortex,
                                                                   path_output_HR=path_output_SR,
                                                                   weights_path=weights_path,
                                                                   by_batch=by_batch,
                                                                   is_conditional=is_conditional
                                                                   )
            else:

                print("already computed")
        else:

            for step in ensemble_pas.loc[patch]:

                i_split = i.split("/")

                path_output = "/".join(i_split[:(len(i_split) - 1)]) + "/code avec " + result_folder + "/patch " + str(
                    patch) + " step " + str(step) + " inversion shave padd"

                print(path_output)

                if not os.path.exists(path_output):

                    create_folder(path_output)

                    path_output_cortex = path_output + "/Cortex patch " + str(patch) + " step " + str(step) + ".nii.gz"

                    path_output_SR = path_output + "/SR.nii.gz" + str(patch) + " step " + str(step) + ".nii.gz"

                    Function_for_application_test_python3.segmentation(input_file_path=i,
                                                                       step=step,
                                                                       NewResolution=(0.5, 0.5, 0.5),
                                                                       patch=patch,
                                                                       path_output_cortex=path_output_cortex,
                                                                       path_output_HR=path_output_SR,
                                                                       weights_path=weights_path,
                                                                       by_batch=by_batch,
                                                                       is_conditional=is_conditional
                                                                       )
                else:

                    print("already computed")



"""
  This software is governed by the CeCILL-B license under French law and
  abiding by the rules of distribution of free software.  You can  use,
  modify and/ or redistribute the software under the terms of the CeCILL-B
  license as circulated by CEA, CNRS and INRIA at the following URL
  "http://www.cecill.info".
  As a counterpart to the access to the source code and  rights to copy,
  modify and redistribute granted by the license, users are provided only
  with a limited warranty  and the software's author,  the holder of the
  economic rights,  and the successive licensors  have only  limited
  liability.
  In this respect, the user's attention is drawn to the risks associated
  with loading,  using,  modifying and/or developing or reproducing the
  software by the user in light of its specific status of free software,
  that may mean  that it is complicated to manipulate,  and  that  also
  therefore means  that it is reserved for developers  and  experienced
  professionals having in-depth computer knowledge. Users are therefore
  encouraged to load and test the software's suitability as regards their
  requirements in conditions enabling the security of their systems and/or
  data to be ensured and,  more generally, to use and operate it in the
  same conditions as regards security.
  The fact that you are presently reading this means that you have had
  knowledge of the CeCILL-B license and that you accept its terms.
"""

import numpy as np
from itertools import product
from sklearn.feature_extraction.image import extract_patches
import SimpleITK as sitk
import scipy.ndimage
import sys
sys.path.insert(0, './utils')
from utils3d import modcrop3D
import os
import shutil
import time
import logging

def array_to_patches(arr, patch_shape=(3,3,3), extraction_step=1, normalization=False):
  #Make use of skleanr function extract_patches
  #https://github.com/scikit-learn/scikit-learn/blob/51a765a/sklearn/feature_extraction/image.py
  """Extracts patches of any n-dimensional array in place using strides.
  Given an n-dimensional array it will return a 2n-dimensional array with
  the first n dimensions indexing patch position and the last n indexing
  the patch content. 
  Parameters
  ----------
  arr : 3darray
      3-dimensional array of which patches are to be extracted
  patch_shape : integer or tuple of length arr.ndim
      Indicates the shape of the patches to be extracted. If an
      integer is given, the shape will be a hypercube of
      sidelength given by its value.
  extraction_step : integer or tuple of length arr.ndim
      Indicates step size at which extraction shall be performed.
      If integer is given, then the step is uniform in all dimensions.
  Returns
  -------
  patches : strided ndarray
      2n-dimensional array indexing patches on first n dimensions and
      containing patches on the last n dimensions. These dimensions
      are fake, but this way no data is copied. A simple reshape invokes
      a copying operation to obtain a list of patches:
      result.reshape([-1] + list(patch_shape))
  """
  
  patches = extract_patches(arr, patch_shape, extraction_step)    
  patches = patches.reshape(-1, patch_shape[0],patch_shape[1],patch_shape[2])    
  # patches = patches.reshape(patches.shape[0], -1) 
  if normalization==True:   
    patches -= np.mean(patches, axis=0)
    patches /= np.std(patches, axis=0)
  print('%.2d patches have been extracted' % patches.shape[0])  ,
  return patches

def patches_to_array(patches, array_shape, patch_shape=(3,3,3) ):
  #Adapted from 2D reconstruction from sklearn
  #https://github.com/scikit-learn/scikit-learn/blob/51a765a/sklearn/feature_extraction/image.py
  # SyntaxError: non-default argument follows default argument : exchange "array_shape" and "patch_shape"
  patches = patches.reshape(len(patches),*patch_shape)
  i_x, i_y, i_z = array_shape
  p_x, p_y, p_z = patch_shape
  array = np.zeros(array_shape)
  # compute the dimensions of the patches array
  n_x = i_x - p_x + 1
  n_y = i_y - p_y + 1
  n_z = i_z - p_z + 1
  for p, (i, j, k) in zip(patches, product(range(n_x), range(n_y), range(n_z))):
      array[i:i + p_x, j:j + p_y, k:k + p_z] += p
  
  for (i, j, k) in product(range(i_x), range(i_y), range(i_z)):
      array[i, j, k] /= float(min(i + 1, p_x, i_x - i) * min(j + 1, p_y, i_y - j) * min(k + 1, p_z, i_z - k))
  return array    

def create_patch_from_df_HR(df,
                            per_cent_val_max,
                            path_save_npy,
                            batch_size,
                            contrast_list,
                            list_res,
                            order=3,
                            normalisation=False,
                            thresholdvalue=0,
                            PatchSize=64,
                            stride=20,
                            is_conditional=False):
    
    Datas_list = []
    
    Labels_list = []
    
    Path_Datas_mini_batch=[]
    
    Path_Labels_mini_batch=[]
    

      
    try:
        
        os.makedirs(path_save_npy)
        print("Dossier cree")
    
    except OSError:
        
        logging.error('Unexpected error: the directory named %s already exists', path_save_npy)
    
    
    mini_batch=0
    
    remaining_patch=0
    
    for  i in range(df.shape[0]):
        
        ReferenceName = df["HR_image"].iloc[i] #path HR
        
        LabelName = df["Label_image"].iloc[i] #path label
        
        print('================================================================')
        print('Processing image : ', ReferenceName)
        
        t1=time.time()
        
        LowResolutionImage , ReferenceImage, LabelImage,UpScale = create_LR_HR_Label(ReferenceName,LabelName,list_res[i])
        
        border_to_keep = border_im_keep(ReferenceImage,thresholdvalue)
        
        ReferenceImage , LowResolutionImage = change_contrast(ReferenceImage,LowResolutionImage,contrast_list[i])
        
        LowResolutionImage = add_noise(LowResolutionImage,per_cent_val_max)
    
        InterpolatedImage , ReferenceImage = norm_and_interp(ReferenceImage,LowResolutionImage,order,UpScale)
        
        LabelImage , ReferenceImage , InterpolatedImage = remove_border(LabelImage,ReferenceImage,InterpolatedImage,border_to_keep) 
        
#        print LabelImage.shape
#        print ReferenceImage.shape
#        print InterpolatedImage.shape
#        print PatchSize
#        print stride
        
        HDF5Labels , HDF5Datas = create_patches(LabelImage,ReferenceImage,InterpolatedImage,PatchSize,stride,normalisation=False)
        
        np.random.seed(0)       # makes the random numbers predictable
        RandomOrder = np.random.permutation(HDF5Datas.shape[0])
        HDF5Datas = HDF5Datas[RandomOrder,:,:,:,:]
        HDF5Labels = HDF5Labels[RandomOrder,:,:,:,:]
        
        if is_conditional : 
            
            #list_res[i] = (res_x,res_y,res_z)
            HDF5Datas = np.concatenate((HDF5Datas,list_res[i][2]*np.ones_like(HDF5Datas)),axis=1) #1st axis = patch axis
        
        Datas_list.append(HDF5Datas)
        
        Labels_list.append(HDF5Labels)
        
        # Tranfer to array
        Datas = np.concatenate(np.asarray(Datas_list))
        Datas = Datas.reshape(-1,
                              Datas.shape[-4],
                              Datas.shape[-3],
                              Datas.shape[-2],
                              Datas.shape[-1]) 
        
        Labels = np.concatenate(np.asarray(Labels_list))
        Labels = Labels.reshape(-1,
                                Labels.shape[-4],
                                Labels.shape[-3],
                                Labels.shape[-2],
                                Labels.shape[-1]) 
                                
        t2=time.time()
        print("Image tranformation + patch creation and organisation :"+str(t2-t1))
                                
                                
        while Datas.shape[0]>=batch_size:   
            
            t1=time.time()
            
            np.save(path_save_npy+"/Datas_mini_batch_"+str(mini_batch)+".npy",Datas[:batch_size,:,:,:,:])
            
            t2=time.time()
            
            print("saving Data array :"+str(t2-t1))
            Datas = Datas[batch_size:,:,:,:,:]
            Datas_list = [Datas]
            
            t1=time.time()
            
            np.save(path_save_npy+"/Label_mini_batch_"+str(mini_batch)+".npy",Labels[:batch_size,:,:,:,:])
            
            t2=time.time()
            
            print("saving Label array :"+str(t2-t1))
            
            Labels = Labels[batch_size:,:,:,:,:]
            Labels_list = [Labels]
            
            
            Path_Datas_mini_batch.append(path_save_npy+"/Datas_mini_batch_"+str(mini_batch)+".npy")
            
            Path_Labels_mini_batch.append(path_save_npy+"/Label_mini_batch_"+str(mini_batch)+".npy")
            
            remaining_patch = Datas.shape[0]
            
            mini_batch+=1
                            
    return path_save_npy,Path_Datas_mini_batch , Path_Labels_mini_batch, remaining_patch  #Label[:,1,:,:,:] : seg , Label[:,0,:,:,:] : HR

def create_patches(Label,HR,interp,PatchSize,stride,normalisation=False):
    
        # Extract 3D patches
        print('Generating training patches ')
        
         
        DataPatch = array_to_patches(interp, 
                                     patch_shape=(PatchSize,PatchSize,PatchSize), 
                                     extraction_step = stride , 
                                     normalization=False) # image interp dim = (nb_patch,patch_size,patch_size,patch_size)
        print('for the interpolated low-resolution patches of training phase.')
                               
        
        LabelHRPatch = array_to_patches(HR, 
                                        patch_shape=(PatchSize,PatchSize,PatchSize), 
                                        extraction_step = stride , 
                                        normalization=False)# image HR dim = (nb_patch,patch_size,patch_size,patch_size)
                                        
        print('for the reference high-resolution patches of training phase.')
                                            
        
        LabelCortexPatch = array_to_patches(Label, 
                                            patch_shape=(PatchSize,PatchSize,PatchSize), 
                                            extraction_step = stride , 
                                            normalization=False) # image seg dim = (nb_patch,patch_size,patch_size,patch_size)
                                            
        print('for the Cortex Labels patches of training phase.') 
        
        # n-dimensional Caffe supports data's form : [numberOfBatches,channels,heigh,width,depth]         
        # Add channel axis !  
        HDF5Datas = DataPatch[:,np.newaxis,:,:,:] #ajoute une dimension de taille 1 en plus 
                
        # Concatenate HR patches and Cortex segmentation : HR patches in the 1st channel and Segmentation the in 2nd channel
        HDF5Labels = np.stack((LabelHRPatch,LabelCortexPatch)) #HDF5Labels[0] = LabelHRPatch et 1=LabelCortexPatch
        
        HDF5Labels = np.swapaxes(HDF5Labels,0,1) #premiere dim = patch ex : HDF5Labels[0,0,:,:,:] = HR premier patch et HDF5Labels[0,1,:,:,:] = Label premier patch
            
        
        return HDF5Labels,HDF5Datas
        
                  
                  
def border_im_keep(HR,thresholdvalue):
    
        darkRegionBox = np.where(HR>thresholdvalue)   
        border = ((np.min(darkRegionBox[0]),np.max(darkRegionBox[0])),
                  (np.min(darkRegionBox[1]),np.max(darkRegionBox[1])),
                  (np.min(darkRegionBox[2]),np.max(darkRegionBox[2])))
                  
        return border
        
def remove_border(Label,HR,interp,border) :


    HR = HR[border[0][0]:border[0][1],border[1][0]:border[1][1],border[2][0]:border[2][1]] 
    Label = Label[border[0][0]:border[0][1],border[1][0]:border[1][1],border[2][0]:border[2][1]]  
    interp = interp[border[0][0]:border[0][1],border[1][0]:border[1][1],border[2][0]:border[2][1]] 
       
    return Label,HR,interp
        
        
def norm_and_interp(HR,LR,order,UpScale):
    # Normalization by the max valeur of LR image
    MaxValue = np.max(LR)
    NormalizedReferenceImage =  HR/MaxValue
    NormalizedLowResolutionImage =  LR/MaxValue
        
    # Cubic Interpolation     
    InterpolatedImage = scipy.ndimage.zoom(NormalizedLowResolutionImage, 
                                  zoom = UpScale,
                                  order = order)
    
    return InterpolatedImage,NormalizedReferenceImage
        
def add_noise(LR,per_cent_val_max):
    
    sigma = per_cent_val_max*np.max(LR)

    LR = LR + np.random.normal(scale=sigma,size=LR.shape)
    
    LR[LR<0]=0

    return LR 

def create_LR_HR_Label(ReferenceName,LabelName,NewResolution):
        
        
        # Read NIFTI
        ReferenceNifti = sitk.ReadImage(ReferenceName)
        
        # Get data from NIFTI
        ReferenceImage = np.swapaxes(sitk.GetArrayFromImage(ReferenceNifti),0,2).astype('float32')
        
                # Read NIFTI
        LabelNifti = sitk.ReadImage(LabelName)
        
        # Get data from NIFTI
        LabelImage = np.swapaxes(sitk.GetArrayFromImage(LabelNifti),0,2).astype('float32')
        
        constant = 2*np.sqrt(2*np.log(2)) # As Greenspan et al. (Full_width_at_half_maximum : slice thickness)
        SigmaBlur = NewResolution/constant  
        
        
        # Get resolution to scaling factor
        UpScale = tuple(itemb/itema for itema,itemb in zip(ReferenceNifti.GetSpacing(),NewResolution))

        # Modcrop to scale factor
        ReferenceImage = modcrop3D(ReferenceImage,UpScale)
        
        LabelImage = modcrop3D(LabelImage,UpScale)
        

        # ===== Generate input LR image =====
        # Blurring
        BlurReferenceImage = scipy.ndimage.filters.gaussian_filter(ReferenceImage,
                                                            sigma = SigmaBlur)
                                                            
        print('Generating LR images with the resolution of ', NewResolution)
      
        # Downsampling
        LowResolutionImage = scipy.ndimage.zoom(BlurReferenceImage,
                                  zoom = (1/float(idxScale) for idxScale in UpScale),
                                  order = 0) 
                                  
        return LowResolutionImage,ReferenceImage,LabelImage,UpScale
        
def change_contrast(HR,LR,power):
    
    HR = HR**power
    LR = LR**power
    
    return HR , LR

        
def create_patch_from_image(ReferenceName,NewResolution,CortexName,PatchSize,batch_size,hdf5name,OutFile,order,thresholdvalue,stride):
    
        constant = 2*np.sqrt(2*np.log(2)) # As Greenspan et al. (Full_width_at_half_maximum : slice thickness)
        SigmaBlur = NewResolution/constant  
        
        print('================================================================')
        print('Processing image : ', ReferenceName)
        # Read NIFTI
        ReferenceNifti = sitk.ReadImage(ReferenceName)
        
        # Get data from NIFTI
        ReferenceImage = np.swapaxes(sitk.GetArrayFromImage(ReferenceNifti),0,2).astype('float32')
        
        # Get resolution to scaling factor
        UpScale = tuple(itemb/itema for itema,itemb in zip(ReferenceNifti.GetSpacing(),NewResolution))

        # Modcrop to scale factor
        ReferenceImage = modcrop3D(ReferenceImage,UpScale)

        # ===== Generate input LR image =====
        # Blurring
        BlurReferenceImage = scipy.ndimage.filters.gaussian_filter(ReferenceImage,
                                                            sigma = SigmaBlur)
                                                            
        print('Generating LR images with the resolution of ', NewResolution)
      
        # Downsampling
        LowResolutionImage = scipy.ndimage.zoom(BlurReferenceImage,
                                  zoom = (1/float(idxScale) for idxScale in UpScale),
                                  order = 0)  
        
        # Normalization by the max valeur of LR image
        MaxValue = np.max(LowResolutionImage)
        NormalizedReferenceImage =  ReferenceImage/MaxValue
        NormalizedLowResolutionImage =  LowResolutionImage/MaxValue
        
        # Cubic Interpolation     
        InterpolatedImage = scipy.ndimage.zoom(NormalizedLowResolutionImage, 
                                  zoom = UpScale,
                                  order = order)  
        
        # Processing cortex segmentation

        print('Processing cortex segmentation map : ', CortexName)
        CortexSegNifti = sitk.ReadImage(CortexName)
        CortexSegmentation = np.swapaxes(sitk.GetArrayFromImage(CortexSegNifti),0,2).astype('float32')
        
        # Shave region outside
        print('Remove the region outside the brain with the value of ', thresholdvalue)
        darkRegionBox = np.where(ReferenceImage>thresholdvalue)   
        border = ((np.min(darkRegionBox[0]),np.max(darkRegionBox[0])),
                  (np.min(darkRegionBox[1]),np.max(darkRegionBox[1])),
                  (np.min(darkRegionBox[2]),np.max(darkRegionBox[2])))     
        LabelHRImage = NormalizedReferenceImage[border[0][0]:border[0][1],border[1][0]:border[1][1],border[2][0]:border[2][1]] 
        LabelCortexImage = CortexSegmentation[border[0][0]:border[0][1],border[1][0]:border[1][1],border[2][0]:border[2][1]]  
        DataImage = InterpolatedImage[border[0][0]:border[0][1],border[1][0]:border[1][1],border[2][0]:border[2][1]]   
        
        # Extract 3D patches
        print('Generating training patches with the resolution of ', NewResolution, ' : ')
        DataPatch = array_to_patches(DataImage, 
                                     patch_shape=(PatchSize,PatchSize,PatchSize), 
                                     extraction_step = stride , 
                                     normalization=False) # image interp dim = (nb_patch,patch_size,patch_size,patch_size)
        print('for the interpolated low-resolution patches of training phase.')                                 
        LabelHRPatch = array_to_patches(LabelHRImage, 
                                        patch_shape=(PatchSize,PatchSize,PatchSize), 
                                        extraction_step = stride , 
                                        normalization=False)# image HR dim = (nb_patch,patch_size,patch_size,patch_size)
        print('for the reference high-resolution patches of training phase.')        
        LabelCortexPatch = array_to_patches(LabelCortexImage, 
                                            patch_shape=(PatchSize,PatchSize,PatchSize), 
                                            extraction_step = stride , 
                                            normalization=False) # image seg dim = (nb_patch,patch_size,patch_size,patch_size)
        print('for the cortex segmentation patches of training phase.')
                          
        # n-dimensional Caffe supports data's form : [numberOfBatches,channels,heigh,width,depth]         
        # Add channel axis !  
        HDF5Datas = DataPatch[:,np.newaxis,:,:,:] #ajoute une dimension de taille 1 en plus 
                
        # Concatenate HR patches and Cortex segmentation : HR patches in the 1st channel and Segmentation the in 2nd channel
        HDF5Labels = np.stack((LabelHRPatch,LabelCortexPatch)) #HDF5Labels[0] = LabelHRPatch et 1=LabelCortexPatch
        HDF5Labels = np.swapaxes(HDF5Labels,0,1) #premiere dim = patch ex : HDF5Labels[0,0,:,:,:] = HR premier patch et HDF5Labels[0,1,:,:,:] = Label premier patch
            
        # Rearrange
        np.random.seed(0)       # makes the random numbers predictable
        RandomOrder = np.random.permutation(HDF5Datas.shape[0])
        HDF5Datas = HDF5Datas[RandomOrder,:,:,:,:]
        HDF5Labels = HDF5Labels[RandomOrder,:,:,:,:]
                   

    
    
    
    
    



"""
  This software is governed by the CeCILL-B license under French law and
  abiding by the rules of distribution of free software.  You can  use,
  modify and/ or redistribute the software under the terms of the CeCILL-B
  license as circulated by CEA, CNRS and INRIA at the following URL
  "http://www.cecill.info".
  As a counterpart to the access to the source code and  rights to copy,
  modify and redistribute granted by the license, users are provided only
  with a limited warranty  and the software's author,  the holder of the
  economic rights,  and the successive licensors  have only  limited
  liability.
  In this respect, the user's attention is drawn to the risks associated
  with loading,  using,  modifying and/or developing or reproducing the
  software by the user in light of its specific status of free software,
  that may mean  that it is complicated to manipulate,  and  that  also
  therefore means  that it is reserved for developers  and  experienced
  professionals having in-depth computer knowledge. Users are therefore
  encouraged to load and test the software's suitability as regards their
  requirements in conditions enabling the security of their systems and/or
  data to be ensured and,  more generally, to use and operate it in the
  same conditions as regards security.
  The fact that you are presently reading this means that you have had
  knowledge of the CeCILL-B license and that you accept its terms.
"""

import numpy as np
import SimpleITK as sitk
import scipy.ndimage
import sys
from ast import literal_eval as make_tuple

sys.path.insert(0, './utils')
from utils3d import modcrop3D
from store2hdf5 import store2hdf53D
from patches import array_to_patches

import argparse

if __name__ == '__main__':

    parser = argparse.ArgumentParser()

    parser.add_argument('-f', '--reference', help='Reference image filename (required)', type=str, action='append', required = True)
    parser.add_argument('-c', '--cortex', help='Cortex segmentation map of reference image (required)', type=str, action='append', required = True)
    parser.add_argument('-o', '--output', help='Name of output HDF5 files (required)', type=str, action='append', required = True)
    parser.add_argument('-n', '--newlowres',  help='Desired low resolution (default = 0.4464,0.4464,3 mm)', type=str, default='0.4464,0.4464,3')
    parser.add_argument('--stride', help='Image extraction stride (default=10)', type=int, default=20)
    parser.add_argument('-b','--batch', help='Indicates batch size for HDF5 storage', type=int, default=16)
    parser.add_argument('--order', help='Order of spline interpolation (default=3) ', type=int, default=3)
    parser.add_argument('-s', '--samples', help='Number of samples of a HDF5 file (optional)', type=int)
    parser.add_argument('-t', '--text', help='Name of a text (.txt) file which contains HDF5 file names (default: train.txt)', type=str, default='train.txt')
    parser.add_argument('--thresholdvalue', help='Value of dark region to remove (default = 0)', type=int, default=0)
    
    args = parser.parse_args()
    
    #  ==== Parser  ===
    # Check number of input and output name:
    if len(args.reference) != len(args.output) :   
        raise AssertionError, 'Number of inputs and outputs should be matched !'    
    
    # Patch size is fixed as 64
    PatchSize = 64
    
    # Check resolution
    NewResolution = make_tuple(args.newlowres)
    if np.isscalar(NewResolution):
        NewResolution = (NewResolution,NewResolution,NewResolution)
    else:
        if len(NewResolution)!=3:
            raise AssertionError, 'Not support this resolution !'  
    
    # Check sigma
    constant = 2*np.sqrt(2*np.log(2)) # As Greenspan et al. (Full_width_at_half_maximum : slice thickness)
    SigmaBlur = NewResolution/constant       
            
    # Writing a text (.txt) file which contains HDF5 file names 
    OutFile = open(str(args.text), "w")
    
    # ============ Processing images ===========================================

    for i in range(0,len(args.reference)):
        # Read reference image
        ReferenceName = args.reference[i]
        print '================================================================'
        print 'Processing image : ', ReferenceName
        # Read NIFTI
        ReferenceNifti = sitk.ReadImage(ReferenceName)
        
        # Get data from NIFTI
        ReferenceImage = np.swapaxes(sitk.GetArrayFromImage(ReferenceNifti),0,2).astype('float32')
        
        # Get resolution to scaling factor
        UpScale = tuple(itemb/itema for itema,itemb in zip(ReferenceNifti.GetSpacing(),NewResolution))

        # Modcrop to scale factor
        ReferenceImage = modcrop3D(ReferenceImage,UpScale)

        # ===== Generate input LR image =====
        # Blurring
        BlurReferenceImage = scipy.ndimage.filters.gaussian_filter(ReferenceImage,
                                                            sigma = SigmaBlur)
                                                            
        print 'Generating LR images with the resolution of ', NewResolution
      
        # Downsampling
        LowResolutionImage = scipy.ndimage.zoom(BlurReferenceImage,
                                  zoom = (1/float(idxScale) for idxScale in UpScale),
                                  order = 0)  
        
        # Normalization by the max valeur of LR image
        MaxValue = np.max(LowResolutionImage)
        NormalizedReferenceImage =  ReferenceImage/MaxValue
        NormalizedLowResolutionImage =  LowResolutionImage/MaxValue
        
        # Cubic Interpolation     
        InterpolatedImage = scipy.ndimage.zoom(NormalizedLowResolutionImage, 
                                  zoom = UpScale,
                                  order = args.order)  
        
        # Processing cortex segmentation
        CortexName = args.cortex[i]
        print 'Processing cortex segmentation map : ', CortexName
        CortexSegNifti = sitk.ReadImage(CortexName)
        CortexSegmentation = np.swapaxes(sitk.GetArrayFromImage(CortexSegNifti),0,2).astype('float32')
        
        # Shave region outside
        print 'Remove the region outside the brain with the value of ', args.thresholdvalue
        darkRegionValue = args.thresholdvalue
        darkRegionBox = np.where(ReferenceImage>darkRegionValue)   
        border = ((np.min(darkRegionBox[0]),np.max(darkRegionBox[0])),
                  (np.min(darkRegionBox[1]),np.max(darkRegionBox[1])),
                  (np.min(darkRegionBox[2]),np.max(darkRegionBox[2])))     
        LabelHRImage = NormalizedReferenceImage[border[0][0]:border[0][1],border[1][0]:border[1][1],border[2][0]:border[2][1]] 
        LabelCortexImage = CortexSegmentation[border[0][0]:border[0][1],border[1][0]:border[1][1],border[2][0]:border[2][1]]  
        DataImage = InterpolatedImage[border[0][0]:border[0][1],border[1][0]:border[1][1],border[2][0]:border[2][1]]   
        
        # Extract 3D patches
        print 'Generating training patches with the resolution of ', NewResolution, ' : '
        DataPatch = array_to_patches(DataImage, 
                                     patch_shape=(PatchSize,PatchSize,PatchSize), 
                                     extraction_step = args.stride , 
                                     normalization=False)
        print 'for the interpolated low-resolution patches of training phase.'                                 
        LabelHRPatch = array_to_patches(LabelHRImage, 
                                        patch_shape=(PatchSize,PatchSize,PatchSize), 
                                        extraction_step = args.stride , 
                                        normalization=False)
        print 'for the reference high-resolution patches of training phase.'        
        LabelCortexPatch = array_to_patches(LabelCortexImage, 
                                            patch_shape=(PatchSize,PatchSize,PatchSize), 
                                            extraction_step = args.stride , 
                                            normalization=False)
        print 'for the cortex segmentation patches of training phase.' 
                          
        # n-dimensional Caffe supports data's form : [numberOfBatches,channels,heigh,width,depth]         
        # Add channel axis !  
        HDF5Datas = DataPatch[:,np.newaxis,:,:,:]
                
        # Concatenate HR patches and Cortex segmentation : HR patches in the 1st channel and Segmentation the in 2nd channel
        HDF5Labels = np.stack((LabelHRPatch,LabelCortexPatch))
        HDF5Labels = np.swapaxes(HDF5Labels,0,1)  
            
        # Rearrange
        np.random.seed(0)       # makes the random numbers predictable
        RandomOrder = np.random.permutation(HDF5Datas.shape[0])
        HDF5Datas = HDF5Datas[RandomOrder,:,:,:,:]
        HDF5Labels = HDF5Labels[RandomOrder,:,:,:,:]
        
        # ============================================================================================
        # Crop data to desired number of samples
        if args.samples :
            HDF5Datas = HDF5Datas[:args.samples ,...]
            HDF5Labels = HDF5Labels[:args.samples ,...]
            
        # Writing to HDF5   
        hdf5name = args.output[i]
        print '*) Writing to HDF5 file : ', hdf5name
        StartLocation = {'dat':(0,0,0,0,0), 'lab': (0,0,0,0,0)}
        CurrentDataLocation = store2hdf53D(filename=hdf5name, 
                                           datas=HDF5Datas, 
                                           labels=HDF5Labels, 
                                           startloc=StartLocation, 
                                           chunksz=args.batch )
                                   
        # Reading HDF5 file                           
        import h5py
        with h5py.File(hdf5name,'r') as hf:
            udata = hf.get('data')
            print 'Shape of input patches:', udata.shape
            ulabel = hf.get('label')
            print 'Shape of output patches:', ulabel.shape

        # Writing a text file which contains HDF5 file names 
        OutFile.write(hdf5name)
        OutFile.write('\n')




"""
  This software is governed by the CeCILL-B license under French law and
  abiding by the rules of distribution of free software.  You can  use,
  modify and/ or redistribute the software under the terms of the CeCILL-B
  license as circulated by CEA, CNRS and INRIA at the following URL
  "http://www.cecill.info".
  As a counterpart to the access to the source code and  rights to copy,
  modify and redistribute granted by the license, users are provided only
  with a limited warranty  and the software's author,  the holder of the
  economic rights,  and the successive licensors  have only  limited
  liability.
  In this respect, the user's attention is drawn to the risks associated
  with loading,  using,  modifying and/or developing or reproducing the
  software by the user in light of its specific status of free software,
  that may mean  that it is complicated to manipulate,  and  that  also
  therefore means  that it is reserved for developers  and  experienced
  professionals having in-depth computer knowledge. Users are therefore
  encouraged to load and test the software's suitability as regards their
  requirements in conditions enabling the security of their systems and/or
  data to be ensured and,  more generally, to use and operate it in the
  same conditions as regards security.
  The fact that you are presently reading this means that you have had
  knowledge of the CeCILL-B license and that you accept its terms.
"""

from keras.legacy import interfaces
import keras.backend as K
from keras.optimizers import Optimizer

class LR_Adam(Optimizer):
    """
    https://ksaluja15.github.io/Learning-Rate-Multipliers-in-Keras/
    
    Adam optimizer.

    Default parameters follow those provided in the original paper.

    # Arguments
        lr: float >= 0. Learning rate.
        beta_1: float, 0 < beta < 1. Generally close to 1.
        beta_2: float, 0 < beta < 1. Generally close to 1.
        epsilon: float >= 0. Fuzz factor.
        decay: float >= 0. Learning rate decay over each update.

    # References
        - [Adam - A Method for Stochastic Optimization](http://arxiv.org/abs/1412.6980v8)
    """

    def __init__(self, lr=0.001, beta_1=0.9, beta_2=0.999,
                 epsilon=1e-8, decay=0., multipliers=None, **kwargs):
        super(LR_Adam, self).__init__(**kwargs)
        with K.name_scope(self.__class__.__name__):
            self.iterations = K.variable(0, name='iterations')
            self.lr = K.variable(lr, name='lr')
            self.beta_1 = K.variable(beta_1, name='beta_1')
            self.beta_2 = K.variable(beta_2, name='beta_2')
            self.decay = K.variable(decay, name='decay')
        self.epsilon = epsilon
        self.initial_decay = decay
        self.lr_multipliers = multipliers              # Multiplier for weights [0,2,4,6,...] and bias [1,3,5,7,...]
        
    @interfaces.legacy_get_updates_support
    def get_updates(self, loss, params):
        grads = self.get_gradients(loss, params)
        self.updates = [K.update_add(self.iterations, 1)]

        lr = self.lr
        if self.initial_decay > 0:
            lr *= (1. / (1. + self.decay * self.iterations))

        t = self.iterations + 1
        lr_t = lr * (K.sqrt(1. - K.pow(self.beta_2, t)) /
                     (1. - K.pow(self.beta_1, t)))

        ms = [K.zeros(K.get_variable_shape(p), dtype=K.dtype(p)) for p in params]
        vs = [K.zeros(K.get_variable_shape(p), dtype=K.dtype(p)) for p in params]
        self.weights = [self.iterations] + ms + vs
        
        # Multiplier for weights [0,2,4,6,...] and bias [1,3,5,7,...]
        if len(params) != len(self.lr_multipliers) : 
            raise "Check Multipliers !" 
        count_multipliers = 0
        
        for p, g, m, v in zip(params, grads, ms, vs):

            # Multiplier for weights [0,2,4,6,...] and bias [1,3,5,7,...]
            if self.lr_multipliers is None:
                new_lr = lr_t     
            else:
                new_lr = lr_t * self.lr_multipliers[count_multipliers]
                count_multipliers += 1
                           
            m_t = (self.beta_1 * m) + (1. - self.beta_1) * g
            v_t = (self.beta_2 * v) + (1. - self.beta_2) * K.square(g)
            p_t = p - new_lr * m_t / (K.sqrt(v_t) + self.epsilon)

            self.updates.append(K.update(m, m_t))
            self.updates.append(K.update(v, v_t))
            new_p = p_t

            # Apply constraints.
            if getattr(p, 'constraint', None) is not None:
                new_p = p.constraint(new_p)

            self.updates.append(K.update(p, new_p))
        return self.updates

    def get_config(self):
        config = {'lr': float(K.get_value(self.lr)),
                  'beta_1': float(K.get_value(self.beta_1)),
                  'beta_2': float(K.get_value(self.beta_2)),
                  'decay': float(K.get_value(self.decay)),
                  'epsilon': self.epsilon}
        base_config = super(LR_Adam, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



"""
  This software is governed by the CeCILL-B license under French law and
  abiding by the rules of distribution of free software.  You can  use,
  modify and/ or redistribute the software under the terms of the CeCILL-B
  license as circulated by CEA, CNRS and INRIA at the following URL
  "http://www.cecill.info".
  As a counterpart to the access to the source code and  rights to copy,
  modify and redistribute granted by the license, users are provided only
  with a limited warranty  and the software's author,  the holder of the
  economic rights,  and the successive licensors  have only  limited
  liability.
  In this respect, the user's attention is drawn to the risks associated
  with loading,  using,  modifying and/or developing or reproducing the
  software by the user in light of its specific status of free software,
  that may mean  that it is complicated to manipulate,  and  that  also
  therefore means  that it is reserved for developers  and  experienced
  professionals having in-depth computer knowledge. Users are therefore
  encouraged to load and test the software's suitability as regards their
  requirements in conditions enabling the security of their systems and/or
  data to be ensured and,  more generally, to use and operate it in the
  same conditions as regards security.
  The fact that you are presently reading this means that you have had
  knowledge of the CeCILL-B license and that you accept its terms.
"""
import numpy as np
import argparse
import os
import sys
sys.path.insert(0, './utils')
from store2hdf5 import ProcessingTrainingSet
from SegSRGAN import SegSRGAN

class SegSRGAN_train(object):
    def __init__(self,
                 TrainingText = 'model/train.txt', 
                 patch=64, 
                 FirstDiscriminatorKernel = 32, FirstGeneratorKernel = 16,
                 lamb_rec = 1, lamb_adv = 0.001, lamb_gp = 10, 
                 lr_DisModel = 0.0001, lr_GenModel = 0.0001):
        
        self.SegSRGAN = SegSRGAN(ImageRow=patch, ImageColumn=patch, ImageDepth=patch, 
                                 FirstDiscriminatorKernel = FirstDiscriminatorKernel, 
                                 FirstGeneratorKernel = FirstGeneratorKernel,
                                 lamb_rec = lamb_rec, lamb_adv = lamb_adv, lamb_gp = lamb_gp, 
                                 lr_DisModel = lr_DisModel, lr_GenModel = lr_GenModel)
        self.generator = self.SegSRGAN.generator()
        self.TrainingText = TrainingText
        self.DiscriminatorModel = self.SegSRGAN.discriminator_model()
        self.GeneratorModel = self.SegSRGAN.generator_model()
        
    def train(self, 
              TrainingEpoch=200, BatchSize=16, SnapshotEpoch=1, InitializeEpoch=1, NumCritic=5, 
              resuming = None):
        
        snapshot_prefix='weights/SegSRGAN_epoch'
        if os.path.exists('weights') is False:
            os.makedirs('weights')
        
        # Initialization Parameters
        real = -np.ones([BatchSize, 1], dtype=np.float32)
        fake = -real
        dummy = np.zeros([BatchSize, 1], dtype=np.float32)   
        
        # Data processing
        TrainingSet = ProcessingTrainingSet(self.TrainingText,BatchSize, InputName='data', LabelName = 'label')

        # Resuming
        if InitializeEpoch==1:
            iteration = 0
            if resuming is None:
                print "Training from scratch"
            else:
                print "Training from the pretrained model (names of layers must be identical): ", resuming
                self.GeneratorModel.load_weights(resuming, by_name=True)
        
        elif InitializeEpoch <1:
            raise AssertionError, 'Resumming needs a positive epoch'
        else:
            if resuming is None:
                raise AssertionError, 'We need pretrained weights'
            else:
                print 'Continue training from : ', resuming
                self.GeneratorModel.load_weights(resuming, by_name=True)
                iteration = (InitializeEpoch-1)*TrainingSet.iterationPerEpoch
        
        # Training phase
        for EpochIndex in range(InitializeEpoch,TrainingEpoch+1):  
            print "Processing epoch : " + str(EpochIndex)
            for iters in range(0,TrainingSet.iterationPerEpoch):
                iteration += 1
                
                # Training discriminator
                for cidx in range(NumCritic):
                    # Loading data randomly
                    randomNumber = int(np.random.randint(0,TrainingSet.iterationPerEpoch,1))
                    DataIndice = randomNumber*BatchSize
                    train_input = TrainingSet.Datas[DataIndice:DataIndice+BatchSize,:,:,:,:]  
                    train_output = TrainingSet.Labels[DataIndice:DataIndice+BatchSize,:,:,:,:]     
                    
                    # Generating fake and interpolation images
                    fake_images = self.generator.predict(train_input)
                    epsilon = np.random.uniform(0, 1, size=(BatchSize,2,1,1,1))
                    interpolation = epsilon*train_output + (1-epsilon)*fake_images

                    # Training
                    dis_loss = self.DiscriminatorModel.train_on_batch([train_output,fake_images,interpolation],
                                                                       [real,fake,dummy])
                    print "Update "+ str(cidx) + ": [D loss: "+str(dis_loss)+"]"  
                    
                # Training generator
                # Loading data
                DataIndice = iters*BatchSize                
                train_input_gen = TrainingSet.Datas[DataIndice:DataIndice+BatchSize,:,:,:,:]  
                train_output_gen = TrainingSet.Labels[DataIndice:DataIndice+BatchSize,:,:,:,:]  

                # Training                                      
                gen_loss = self.GeneratorModel.train_on_batch([train_input_gen], 
                                                               [real,train_output_gen])
                print "Iter "+ str(iteration) + " [A loss: " + str(gen_loss) + "]"  
                
            if (EpochIndex)%SnapshotEpoch==0:
                # Save weights:
                self.GeneratorModel.save_weights(snapshot_prefix + '_' + str(EpochIndex))
                print ("Snapshot :" + snapshot_prefix + '_' + str(EpochIndex))
                          


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    
    parser.add_argument('-t', '--text', help='Name of a text (.txt) file which contains HDF5 file names (default: train.txt)', type=str, default='train.txt')
    parser.add_argument('-e', '--epoch', help='Number of training epochs (default=200)', type=int, default=200)
    parser.add_argument('-b', '--batchsize', help='Number of batch (default=16)', type=int, default=16)
    parser.add_argument('-s', '--snapshot', help='Snapshot Epoch (default=1)', type=int, default=1)
    parser.add_argument('-i', '--initepoch', help='Init Epoch (default=1)', type=int, default=1)
    parser.add_argument('-w', '--weights', help='Name of the pretrained HDF5 weight file (default: None)', type=str, default=None)
    parser.add_argument('--kernelgen', help='Number of filters of the first layer of generator (default=16)', type=int, default=16)
    parser.add_argument('--kerneldis', help='Number of filters of the first layer of discriminator (default=32)', type=int, default=32)
    parser.add_argument('--lrgen', help='Learning rate of generator (default=0.0001)', type=int, default=0.0001)
    parser.add_argument('--lrdis', help='Learning rate of discriminator (default=0.0001)', type=int, default=0.0001)
    parser.add_argument('--lambrec', help='Lambda of reconstruction loss (default=1)', type=int, default=1)
    parser.add_argument('--lambadv', help='Lambda of adversarial loss (default=0.001)', type=int, default=0.001)
    parser.add_argument('--lambgp', help='Lambda of gradien penalty loss (default=10)', type=int, default=10)
    parser.add_argument('--numcritic', help='Number of training time for discriminator (default=5) ', type=int, default=5)
    
    args = parser.parse_args()
    SegSRGAN_train = SegSRGAN_train(TrainingText = args.text,  
                                    FirstDiscriminatorKernel = args.kerneldis, FirstGeneratorKernel = args.kernelgen,
                                    lamb_rec = args.lambrec, lamb_adv = args.lambadv, lamb_gp = args.lambgp, 
                                    lr_DisModel = args.lrdis, lr_GenModel = args.lrgen)
    SegSRGAN_train.train(TrainingEpoch=args.epoch, BatchSize=args.batchsize, 
                         SnapshotEpoch=args.snapshot, InitializeEpoch = args.initepoch,
                         NumCritic = args.numcritic,
                         resuming = args.weights)



"""
  This software is governed by the CeCILL-B license under French law and
  abiding by the rules of distribution of free software.  You can  use,
  modify and/ or redistribute the software under the terms of the CeCILL-B
  license as circulated by CEA, CNRS and INRIA at the following URL
  "http://www.cecill.info".
  As a counterpart to the access to the source code and  rights to copy,
  modify and redistribute granted by the license, users are provided only
  with a limited warranty  and the software's author,  the holder of the
  economic rights,  and the successive licensors  have only  limited
  liability.
  In this respect, the user's attention is drawn to the risks associated
  with loading,  using,  modifying and/or developing or reproducing the
  software by the user in light of its specific status of free software,
  that may mean  that it is complicated to manipulate,  and  that  also
  therefore means  that it is reserved for developers  and  experienced
  professionals having in-depth computer knowledge. Users are therefore
  encouraged to load and test the software's suitability as regards their
  requirements in conditions enabling the security of their systems and/or
  data to be ensured and,  more generally, to use and operate it in the
  same conditions as regards security.
  The fact that you are presently reading this means that you have had
  knowledge of the CeCILL-B license and that you accept its terms.
"""
import numpy as np
import argparse
import os
import sys
sys.path.insert(0, './utils')
from SegSRGAN import SegSRGAN
from patches import create_patch_from_df_HR
import pandas as pd
from ast import literal_eval as make_tuple
import shutil
import time
     

class SegSRGAN_train(object):
    def __init__(self,
                 BasePath,
                 contrast_max,
                 percent_val_max,
                 list_res_max,
                 Trainingcsv, 
                 multi_gpu,
                 patch=64, 
                 FirstDiscriminatorKernel = 32, FirstGeneratorKernel = 16,
                 lamb_rec = 1, lamb_adv = 0.001, lamb_gp = 10, 
                 lr_DisModel = 0.0001, lr_GenModel = 0.0001,u_net_gen=False,
                 is_conditional=False,
                 is_residual=True):
        
        self.SegSRGAN = SegSRGAN(ImageRow=patch, ImageColumn=patch, ImageDepth=patch, 
                                 FirstDiscriminatorKernel = FirstDiscriminatorKernel, 
                                 FirstGeneratorKernel = FirstGeneratorKernel,
                                 lamb_rec = lamb_rec, lamb_adv = lamb_adv, lamb_gp = lamb_gp, 
                                 lr_DisModel = lr_DisModel, lr_GenModel = lr_GenModel,u_net_gen=u_net_gen,multi_gpu=multi_gpu,is_conditional=is_conditional)
        self.generator = self.SegSRGAN.generator()
        self.Trainingcsv = Trainingcsv
        self.DiscriminatorModel, self.DiscriminatorModel_multi_gpu = self.SegSRGAN.discriminator_model()
        self.GeneratorModel, self.GeneratorModel_multi_gpu = self.SegSRGAN.generator_model()
        self.BasePath = BasePath
        self.contrast_max = contrast_max
        self.percent_val_max = percent_val_max
        self.list_res_max = list_res_max
        self.multi_gpu=multi_gpu
        self.is_conditional=is_conditional
        self.is_residual=is_residual
        
        print("initalisation finised")
        
    def train(self,
              snapshot_folder,
              dice_file,
              mse_file, 
              folder_training_data,
              TrainingEpoch=200, BatchSize=16, SnapshotEpoch=1, InitializeEpoch=1, NumCritic=5, 
              resuming = None):
        
        #snapshot_prefix='weights/SegSRGAN_epoch'
        print("train begin")
        snapshot_prefix = snapshot_folder+"/SegSRGAN_epoch"
        
        # boolean to print only one time 'the number of patch not in one epoche (mod batchsize)'
        never_print = True
        if os.path.exists(snapshot_folder) is False:
            os.makedirs(snapshot_folder)
        
        # Initialization Parameters
        real = -np.ones([BatchSize, 1], dtype=np.float32)
        fake = -real
        dummy = np.zeros([BatchSize, 1], dtype=np.float32)   
        
        # Data processing
        #TrainingSet = ProcessingTrainingSet(self.TrainingText,BatchSize, InputName='data', LabelName = 'label')
        
        data=pd.read_csv(self.Trainingcsv)
        
        data["HR_image"] = self.BasePath+data["HR_image"]
        data["Label_image"] = self.BasePath+data["Label_image"]

        
        data_train =data[data['Base']=="Train"]
        data_test = data[data['Base']=="Test"]
        
        # Resuming
        if InitializeEpoch==1:
            iteration = 0
            if resuming is None:
                print ("Training from scratch")
            else:
                print("Training from the pretrained model (names of layers must be identical): ", resuming)
                self.GeneratorModel.load_weights(resuming, by_name=True)
        
        elif InitializeEpoch <1:
            raise AssertionError('Resumming needs a positive epoch')
        else:
            if resuming is None:
                raise AssertionError('We need pretrained weights')
            else:
                print('Continue training from : ', resuming)
                self.GeneratorModel.load_weights(resuming, by_name=True)
#                iteration = (InitializeEpoch-1)*iterationPerEpoch
        # patch test creation :
                
        t1=time.time()
        
        test_contrast_list=np.linspace(1-self.contrast_max,1+self.contrast_max,data_test.shape[0])
        
        
        # list_res[0] = lower bound and list_res[1] = borne supp
        # list_res[0][0] = lower bound for the first coordinate
        
        lin_res_x = np.linspace(self.list_res_max[0][0],self.list_res_max[1][0],data_test.shape[0])
        lin_res_y = np.linspace(self.list_res_max[0][1],self.list_res_max[1][1],data_test.shape[0])
        lin_res_z = np.linspace(self.list_res_max[0][2],self.list_res_max[1][2],data_test.shape[0])
        
          
        res_test = [(lin_res_x[i],
                     lin_res_y[i],
                     lin_res_z[i]) for i in range(data_test.shape[0])]
                
        test_path_save_npy,test_Path_Datas_mini_batch , test_Labels_mini_batch, test_remaining_patch = create_patch_from_df_HR(df = data_test,
                                                                                                         per_cent_val_max = self.percent_val_max,
                                                                                                         contrast_list = test_contrast_list,
                                                                                                         list_res = res_test,
                                                                                                         order=3,
                                                                                                         normalisation=False,
                                                                                                         thresholdvalue=0,
                                                                                                         PatchSize=64,
                                                                                                         batch_size = 1, #1 to keep all data
                                                                                                         path_save_npy=folder_training_data+"/test_mini_batch",
                                                                                                         stride=20,is_conditional=self.is_conditional)
        
        t2=time.time()
        
        print("time for making test npy :"+str(t2-t1))   

        df_dice = pd.DataFrame(index=np.arange(InitializeEpoch,TrainingEpoch+1),columns=["Dice"])
        df_MSE = pd.DataFrame(index=np.arange(InitializeEpoch,TrainingEpoch+1),columns=["MSE"])
        
        # Training phase
        for EpochIndex in range(InitializeEpoch,TrainingEpoch+1):  
    
            train_contrast_list=np.random.uniform(1-self.contrast_max,1+self.contrast_max,data_train.shape[0])
            
            
            res_train = [(np.random.uniform(self.list_res_max[0][0],self.list_res_max[1][0]),
                          np.random.uniform(self.list_res_max[0][1],self.list_res_max[1][1]),
                          np.random.uniform(self.list_res_max[0][2],self.list_res_max[1][2])) for i in range(data_train.shape[0])]
            
            t1=time.time()
        
            train_path_save_npy,train_Path_Datas_mini_batch ,train_Labels_mini_batch, train_remaining_patch = create_patch_from_df_HR  (df = data_train,
                                                                                                               per_cent_val_max = self.percent_val_max,
                                                                                                               contrast_list = train_contrast_list,
                                                                                                               list_res = res_train,
                                                                                                               order=3,
                                                                                                               normalisation=False,
                                                                                                               thresholdvalue=0,
                                                                                                               PatchSize=64,
                                                                                                               batch_size = BatchSize,
                                                                                                               path_save_npy=folder_training_data+"/train_mini_batch",
                                                                                                               stride=20,
                                                                                                               is_conditional=self.is_conditional)
            iterationPerEpoch = len(train_Path_Datas_mini_batch)
            
            t2=time.time()
        
            print("time for making train npy :"+str(t2-t1))

            if never_print: 
                                                                                                 
                print("At each epoch "+str(train_remaining_patch)+" patches will not be in the training data for this epoch")
                never_print = False                                          
                                                      
            
            print("Processing epoch : " + str(EpochIndex))
            for iters in range(0,iterationPerEpoch):
                
                iteration += 1
                
                # Training discriminator
                for cidx in range(NumCritic):
                    
                    t1=time.time()

                    # Loading data randomly
                    randomNumber = int(np.random.randint(0,iterationPerEpoch,1))
                    
                    train_input = np.load(train_Path_Datas_mini_batch[randomNumber])[:,0,:,:,:][:,np.newaxis,:,:,:]# select 0 coordoniate and add one axis at the same place
                    
                    train_output = np.load(train_Labels_mini_batch[randomNumber])                    
                    
                    if self.is_conditional :
                        
                        train_res = np.load(train_Path_Datas_mini_batch[randomNumber])[:,1,:,:,:][:,np.newaxis,:,:,:]
                         
                        # Generating fake and interpolation images
                        fake_images = self.GeneratorModel_multi_gpu.predict([train_input,train_res])[1]
                        epsilon = np.random.uniform(0, 1, size=(BatchSize,2,1,1,1))
                        interpolation = epsilon*train_output + (1-epsilon)*fake_images
                        # Training
                        dis_loss = self.DiscriminatorModel_multi_gpu.train_on_batch([train_output,fake_images,interpolation,train_res],
                                                                           [real,fake,dummy])
                    else :
                        
                        # Generating fake and interpolation images
                        fake_images = self.GeneratorModel_multi_gpu.predict(train_input)[1]
                        epsilon = np.random.uniform(0, 1, size=(BatchSize,2,1,1,1))
                        interpolation = epsilon*train_output + (1-epsilon)*fake_images
                        # Training
                        dis_loss = self.DiscriminatorModel_multi_gpu.train_on_batch([train_output,fake_images,interpolation],
                                                                           [real,fake,dummy])
                         
                    t2=time.time()
        
                    print("time for one uptade of discriminator :"+str(t2-t1))
                    print("Update "+ str(cidx) + ": [D loss : "+str(dis_loss)+"]") 
                    
                # Training generator
                # Loading data        
                
                t1=time.time()
                
                train_input_gen = np.load(train_Path_Datas_mini_batch[iters])[:,0,:,:,:][:,np.newaxis,:,:,:]
                train_output_gen = np.load(train_Labels_mini_batch[iters])
                
                if self.is_conditional:
                    
                    train_res_gen = np.load(train_Path_Datas_mini_batch[iters])[:,1,:,:,:][:,np.newaxis,:,:,:]
                    # Training                                      
                    gen_loss = self.GeneratorModel_multi_gpu.train_on_batch([train_input_gen,train_res_gen], 
                                                                   [real,train_output_gen])
                else : 
                    # Training                                      
                    gen_loss = self.GeneratorModel_multi_gpu.train_on_batch([train_input_gen], 
                                                                   [real,train_output_gen])
                    
                                                                       
                print("Iter "+ str(iteration) + " [A loss : " + str(gen_loss) + "]")
                
                t2=time.time()
                
                print("time for one uptade of generator :"+str(t2-t1))
                
            if (EpochIndex)%SnapshotEpoch==0:
                # Save weights:
                self.GeneratorModel.save_weights(snapshot_prefix + '_' + str(EpochIndex))
                print ("Snapshot :" + snapshot_prefix + '_' + str(EpochIndex))
                
            MSE_list= []
            VP = []
            Pos_pred = []
            Pos_label = []
            
            t1 = time.time()
            
            
            for test_iter in range(len(test_Labels_mini_batch)):

                TestLabels = np.load(test_Labels_mini_batch[test_iter])
                TestDatas = np.load(test_Path_Datas_mini_batch[test_iter])[:,0,:,:,:][:,np.newaxis,:,:,:]
                
                if self.is_conditional : 
                    
                    TestRes = np.load(test_Path_Datas_mini_batch[test_iter])[:,1,:,:,:][:,np.newaxis,:,:,:]
                    
                    pred = self.generator.predict([TestDatas,TestRes])
                    
                else : 
                    
                    pred = self.generator.predict([TestDatas])
                    
                

                pred[:,0,:,:,:][pred[:,0,:,:,:]<0]=0            
            
                MSE_list.append(np.sum((pred[:,0,:,:,:]-TestLabels[:,0,:,:,:])**2))
                
                VP.append(np.sum((pred[:,1,:,:,:]>0.5)&(TestLabels[:,1,:,:,:]==1)))
                
                Pos_pred.append(np.sum(pred[:,1,:,:,:]>0.5))
                
                Pos_label.append(np.sum(TestLabels[:,1,:,:,:]))
                
            t2 = time.time()
            
            print("Evaluation on test data time : " + str(t2-t1))
                
            gen_weights = np.array(self.GeneratorModel.get_weights())
            gen_weights_multi = np.array(self.GeneratorModel_multi_gpu.get_weights())
            
            weights_idem =True
            
            for i in range(len(gen_weights)) :
                
                idem =  np.array_equal(gen_weights[i],gen_weights_multi[i])
                
                weights_idem = weights_idem & idem 
                
            if weights_idem : 
                
                print("Model multi_gpu and base Model have the same weights")
                
            else :
                print("Model multi_gpu and base Model haven't the same weights")
                
            
            Dice = (2 * np.sum(VP))/(np.sum(Pos_pred)+np.sum(Pos_label))
            
            MSE = np.sum(MSE_list)/(BatchSize**3*len(MSE_list)) 
            
            print("Iter "+ str(EpochIndex) + " [Test Dice : " + str(Dice) + "]") 
            
            print("Iter "+ str(EpochIndex) + " [Test MSE : " + str(MSE) + "]")  
            
            df_dice.loc[EpochIndex,"Dice"]=Dice
            df_MSE.loc[EpochIndex,"MSE"]=MSE
            
            df_dice.to_csv(dice_file)
            df_MSE.to_csv(mse_file)
            
            shutil.rmtree(folder_training_data+"/train_mini_batch")
        
        shutil.rmtree(folder_training_data+"/test_mini_batch")
    
            
            
                        

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('-begining_path', '--basepath', help='path to concatenate with relative path contains in csv file ', type=str, required = True) # path qui apres concatenation des path contenu dans le fichier csv amene au fichier.
    parser.add_argument('-n', '--newlowres', type=float, action='append',help='upper and lower bounds between which the low resolution of each image at each epoch will be choosen randomly' ,nargs="+",required=True)
    parser.add_argument('-contrast_max', '--contrast_max', help='Ex : 0.3 : NN trained on contrast between power 0.3 and 1.3 of initial image (default=0.5)', type=float,default=0.5)
    parser.add_argument('-percent_val_max', '--percent_val_max', help='NN trained on image on which we add gaussian noise with sigma equal this % of val_max', type=float,default=0.03)
    parser.add_argument('-csv', '--csv', help='.csv contining relative path for testing and training base', type=str, required = True) # tous les path se tranvant dans le fichier sont relatif a begining_path, collone HR_image : path HR Label_image : path Label	mask : path mask Base : "Train" ou "Test",
    parser.add_argument('-sf', '--snapshot_folder', help='Folder name for saving spanshot weights', type=str, required = True)
    parser.add_argument('-e', '--epoch', help='Number of training epochs (default=200)', type=int, default=200)
    parser.add_argument('-b', '--batchsize', help='Number of batch (default=16)', type=int, default=16)
    parser.add_argument('-s', '--snapshot', help='Snapshot Epoch (default=1)', type=int, default=1)
    parser.add_argument('-i', '--initepoch', help='Init Epoch (default=1)', type=int, default=1)
    parser.add_argument('-w', '--weights', help='Name of the pretrained HDF5 weight file (default: None)', type=str, default=None)
    parser.add_argument('--kernelgen', help='Number of filters of the first layer of generator (default=16)', type=int, default=16)
    parser.add_argument('--kerneldis', help='Number of filters of the first layer of discriminator (default=32)', type=int, default=32)
    parser.add_argument('--lrgen', help='Learning rate of generator (default=0.0001)', type=int, default=0.0001)
    parser.add_argument('--lrdis', help='Learning rate of discriminator (default=0.0001)', type=int, default=0.0001)
    parser.add_argument('--lambrec', help='Lambda of reconstruction loss (default=1)', type=int, default=1)
    parser.add_argument('--lambadv', help='Lambda of adversarial loss (default=0.001)', type=int, default=0.001)
    parser.add_argument('--lambgp', help='Lambda of gradien penalty loss (default=10)', type=int, default=10)
    parser.add_argument('--numcritic', help='Number of training time for discriminator (default=5) ', type=int, default=5)
    parser.add_argument('-dice','--dice_file', help='Dice path for save dice a the end of each epoche', type=str,required = True)
    parser.add_argument('-mse','--mse_file', help='MSE path for save dice a the end of each epoche', type=str,required = True)
    parser.add_argument('-u_net','--u_net_generator', help='Either the generator take u-net architecture (like u-net) or not. Value in {True,False} default : False', type=str,default="False")
    parser.add_argument('-folder_training_data','--folder_training_data', help="folder in which data organized by batch will be save during training (this folder will be created)", type=str,required =True)
    parser.add_argument('-multi_gpu','--multi_gpu', help="Train using all gpu available if some exist ? Value in {True,False} default : True", type=str,default="True")
    parser.add_argument('-is_conditional','--is_conditional', help="Should a conditionnal GAN be train ? Value in {True,False} default : False", type=str,default="False")
    parser.add_argument('-is_residual','--is_residual', help="Should a residual GAN be train (sum of pred and image for SR estimation) ? Value in {True,False} default : True", type=str,default="True")
    args = parser.parse_args()
    
    
    u_net = (args.u_net_generator=="True") #Transform str to boolean
    
    multi_gpu = (args.multi_gpu=="True")
    
    is_residual = (args.is_residual=="True")
    
    is_conditional = (args.is_conditional=="True")
    
    print("percent val max :"+str(args.percent_val_max))
    
    print("u_net = "+str(u_net))
    
    print("is_conditional = "+str(is_conditional))
    
    print("is_residual = "+str(is_residual))
    
    
    list_res_max = args.newlowres
    
    
    for i in range(len(list_res_max)):
        
        if len(list_res_max[i])!=3:
            raise AssertionError('Not support this resolution !')
            
    print("Initial resolution given "+str(list_res_max))
                
            
    if len(list_res_max)==1:
        
        list_res_max.extend(list_res_max)
        
    print("the low resolution of images will be choosen randomly between "+str(list_res_max[0])+" and "+str(list_res_max[1]))
        
        

    SegSRGAN_train = SegSRGAN_train(Trainingcsv = args.csv,  
                                    contrast_max=args.contrast_max,
                                    percent_val_max=args.percent_val_max,
                                    FirstDiscriminatorKernel = args.kerneldis, FirstGeneratorKernel = args.kernelgen,
                                    lamb_rec = args.lambrec, lamb_adv = args.lambadv, lamb_gp = args.lambgp, 
                                    lr_DisModel = args.lrdis, lr_GenModel = args.lrgen,BasePath=args.basepath,list_res_max=list_res_max,u_net_gen=u_net,multi_gpu=multi_gpu,is_conditional=is_conditional,is_residual=is_residual)
                                    
    SegSRGAN_train.train(TrainingEpoch=args.epoch, BatchSize=args.batchsize, 
                         SnapshotEpoch=args.snapshot, InitializeEpoch = args.initepoch,
                         NumCritic = args.numcritic,
                         resuming = args.weights,
                         dice_file=args.dice_file,
                         mse_file=args.mse_file,
                         snapshot_folder=args.snapshot_folder,
                         folder_training_data = args.folder_training_data
                         )
                         
                         


import setuptools
from os.path import abspath

with open(abspath("SegSRGAN/README.md"), "r") as fh:
    long_description = fh.read()

with open(abspath("SegSRGAN/requirements.txt")) as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="SegSRGAN",
    version="1.0.4",
    author="Clément Cazorla",
    author_email="clement.cazorla@univ-reims.fr",
    description="Segmentation and super resolution GAN network",
    long_description=long_description,
    long_description_content_type="text/plain",
    url="https://github.com/koopa31/SegSRGAN/tree/develop",
    packages=setuptools.find_packages(),
    install_requires=requirements,
)
